# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Luke Macken <lmacken@redhat.com>

import logging
import feedparser
import pkg_resources
import time
import os

try:
    from cStringIO import cStringIO as StringIO
except ImportError:
    from StringIO import StringIO

from twisted.web import client
from twisted.web.client import HTTPPageGetter, HTTPClientFactory
from twisted.internet import reactor, protocol, defer
from paste.deploy.converters import asbool

from datetime import timedelta
from feedcache import Cache
from shove import Shove
from tg import config

from moksha.hub.hub import MokshaHub
from moksha.api.hub.producer import PollingProducer

log = logging.getLogger('moksha.hub')

feed_storage = Shove(config.get('feed_cache', 'simple://'), compress=True)
feed_cache = Cache(feed_storage)

class ConditionalHTTPPageGetter(HTTPPageGetter):

    def handleStatus_200(self):
        """ Attempt to save the last-modified header """
        if self.headers.has_key('last-modified'):
            self.factory.lastModified(self.headers['last-modified'])

    def handleStatus_304(self):
        """ Close the connection """
        self.factory.notModified()
        self.transport.loseConnection()


class ConditionalHTTPClientFactory(HTTPClientFactory):

    protocol = ConditionalHTTPPageGetter

    def __init__(self, url, method='GET', postdata=None, headers=None,
                 agent=None, timeout=0, cookies=None, followRedirect=1):

        self.url = url

        try:
            if url in feed_storage:
                lastModified = time.ctime(feed_storage[url][0])
                if headers is not None:
                    headers['last-modified'] = lastModified
                else:
                    headers = {'last-modified': lastModified}
        except Exception, e:
            log.error("Unable to iterate over feed_storage")

        HTTPClientFactory.__init__(self, url, method=method, postdata=postdata,
                headers=headers, agent=agent, timeout=timeout, cookies=cookies,
                followRedirect=followRedirect)

        self.waiting = True
        self.deferred = defer.Deferred()

    def lastModified(self, modtime):
        try:
            t = time.mktime(time.strptime(modtime[0], '%a, %d %b %Y %H:%M:%S %Z'))
        except ValueError:
            # Try stripping off the timezone?
            t = time.mktime(time.strptime(' '.join(modtime[0].split()[:-1]),
                                          '%a, %d %b %Y %H:%M:%S'))

        parsed_feed = {}

        if self.url in feed_storage:
            current_feed = feed_storage[self.url][1]
            if current_feed and not current_feed.get('bozo_exception'):
                parsed_feed = current_feed

        try:
            feed_storage[self.url] = (t, parsed_feed)
            print feed_storage.sync.__doc__
            feed_storage.sync()
        except Exception, e:
            log.error("Unable to store parsed_feed: %r" % parsed_feed)
            log.exception(e)

    def notModified(self):
        if self.waiting:
            self.waiting = False


def conditional_get_page(url, contextFactory=None, *args, **kwargs):
    scheme, host, port, path = client._parse(url)
    factory = ConditionalHTTPClientFactory(url, *args, **kwargs)
    if scheme == 'https':
        from twisted.internet import ssl
        if contextFactory is None:
            contextFactory = ssl.ClientContextFactory()
        reactor.connectSSL(host, port, factory, contextFactory)
    else:
        reactor.connectTCP(host, port, factory)
    return factory.deferred


class FeederProtocol(object):

    max_age = int(config.get('feed.max_age', 300))
    timeout = int(config.get('feed.timeout', 60))

    def __init__(self):
        self.parsed = 1
        self.hub = MokshaHub()
        self.post_processors = []
        for entry in pkg_resources.iter_entry_points('moksha.feeds.post_processor'):
            log.info('Registering feed post-processor: %s' % entry.name)
            self.post_processors.append(entry.load())

    def is_cached(self, site):
        already_got = feed_storage.get(site)
        if already_got:
            elapsed_time = time.time() - already_got[0]
            if elapsed_time < self.max_age:
                return True
            else:
                return False
        else:
            return False

    def on_error(self, traceback, extra_args):
        log.error(extra_args)
        log.exception(traceback)

    def get_feed_from_cache(self, data, key=None):
        """ Return feed data from the cache based on a given ``key`` """
        log.debug('Getting cached feed for %s' % key)
        return defer.succeed(feed_storage.get(key, key)[1])

    def parse_feed(self, feed, url):
        if not feed:
            log.warning('parse_feed got %r for %s' % (feed, url))
            return {}
        if not isinstance(feed, basestring):
            feed = str(feed)
        feed = feedparser.parse(StringIO(feed))
        assert feed
        if feed.get('bozo_exception'):
            bozo_exc = str(feed['bozo_exception'])
            log.warning("Feed %s getting bozo_exception %r" % (feed, bozo_exc))
            feed['bozo_exception'] = bozo_exc
        return feed

    def store_feed(self, feed, addr):
        try:
            feed_storage[addr] = (time.time(), feed)
        except Exception, e:
            log.error('Unable to store feed %s: %s' % (addr, str(e)))
        return feed

    def get_feed(self, addr):
        try:
            return feed_storage[addr][1]
        except KeyError:
            return None

    def process_feed(self, parsed_feed, addr, olddata):
        """ Process the parsed feed.

        If `olddata` is provided, this method will look for new feed entries,
        and send notifications to the `feed.$FEED_URL` MokshaHub Topic.

        :param parsed_feed: A parsed :mod:`feedcache` feed
        :param addr: The URL of the feed
        :param olddata: The cached feed data
        """
        if not parsed_feed:
            log.error("Cannot process %r feed for %s" % (parsed_feed, addr))
            return

        chan = parsed_feed.get('channel', None)
        if chan:
            log.debug(chan.get('title', ''))

        # Previous data provided; look for new entries.
        if olddata:
            oldtitles = [entry.get('title') for entry in olddata['entries']]
            new_entries = parsed_feed.get('entries', [{}])
            if not len(new_entries):
                log.warning('Feed contains empty entries: %s' % addr)
                return

            # If there are no new entries, move on...
            newtitle = new_entries[0].get('title', None)
            if newtitle == oldtitles[0]:
                return

            # Send notifications for each new entry
            for entry in new_entries[::-1]:
                entry_title = entry.get('title', '[No Title]')
                channel_link = entry.get('channel', {'link': addr})['link']
                if entry['title'] not in oldtitles:
                    log.info('New feed entry found: %s' % entry['title'])
                    if self.post_processors:
                        for processor in self.post_processors:
                            entry = processor(entry)
                        try:
                            self.hub.send_message('moksha.feeds.%s' % channel_link, entry)
                        except Exception, e: # Usually JSON encoding issues.
                            log.error(str(e))
                            log.debug('Sending just the title and link instead')
                            self.hub.send_message('moksha.feeds.%s' % channel_link,
                                    {'title': entry_title, 'link': entry.get('link')})
                    else:
                        self.hub.send_message('moksha.feeds.%s' % channel_link,
                                {'title': entry_title, 'link': entry.get('link')})

    def get_page(self, data, args):
        return conditional_get_page(args, timeout=self.timeout)

    def start(self, data=None):
        d = defer.succeed(True)
        for feed in data:
            olddata = None
            if self.is_cached(feed):
                d.addCallback(self.get_feed_from_cache, feed)
                d.addErrback(self.on_error, (feed, 'fetching from cache'))
            else:
                d.addCallback(self.get_page, feed)
                d.addErrback(self.on_error, (feed, 'fetching'))
                d.addCallback(self.parse_feed, feed)
                d.addErrback(self.on_error, (feed, 'parsing'))
                olddata = self.get_feed(feed)
                d.addCallback(self.store_feed, feed)
                d.addErrback(self.on_error, (feed, 'storing'))
            d.addCallback(self.process_feed, feed, olddata)
            d.addErrback(self.on_error, (feed, 'processing'))
            del(olddata)
        return d


class FeederFactory(protocol.ClientFactory):

    protocol = FeederProtocol()

    def __init__(self):
        """Initialize the Feeder Factory.

        :param deferred_groups: The number of simultaneous connections
        """
        self.protocol.factory = self
        self.deferred_groups = int(config.get('feed.deferred_groups', 50))

    def start(self, addresses):
        """Divide into groups all the feeds to download.

        :param addresses: A list of feed urls
        """
        log.info("Starting the FeederFactory...")
        if len(addresses) > self.deferred_groups:
            url_groups = [[] for x in xrange(self.deferred_groups)]
            for i, addr in enumerate(addresses):
                url_groups[i % self.deferred_groups].append(addr)
        else:
            url_groups = [[addr] for addr in addresses]
        log.info("Creating %d url groups" % len(url_groups))
        for group in url_groups:
            self.protocol.start(group)


class MokshaFeedStream(PollingProducer):
    """
    If you expose your feed widget on the moksha.widget entry point,
    then Moksha will automatically handle polling it.  Upon new entries,
    AMQP messages will be sent to the `feeds.$URL` queue.
    """
    #frequency = timedelta(minutes=1)
    now = False

    def __init__(self, hub):
        enabled = asbool(config.get('moksha.feedaggregator', False))
        if not enabled:
            log.info('Moksha Feed Aggregator disabled')
            return
        else:
            self.frequency = int(config.get('feed.poll_frequency', 900))
        super(MokshaFeedStream, self).__init__(hub)

    def poll(self):
        """ Poll all feeds in our feed cache """
        log.debug('FeedStream.poll()')

        feeds = set()
        for feed in feed_storage.keys():
            feeds.add(str(feed))

        # Read in all feeds from the `feeds.txt` file for testing...
        if os.path.isfile('feeds.txt'):
            feed_list = file('feeds.txt')
            for feed in feed_list.readlines():
                feeds.add(str(feed.strip()))

        f = FeederFactory()
        f.start(addresses=feeds)

    def stop(self):
        feed_storage.close()
        super(MokshaFeedStream, self).stop()



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`moksha.widgets.feedtree.consumer` - The feed consumer
===========================================================

The :class:`MokshaFeedConsumer` listens to all messages on the ``feeds`` topic.
If the ``action`` entry in the message header contains a valid command, it will
execute it.  Most of the time it will perform a specific action, such as getting
entries for a specific feed, and will send the data to the specified ``topic``.

.. moduleauthor:: Luke Macken <lmacken@redhat.com>
"""

from tg import config
from shove import Shove
from kitchen.text.converters import to_unicode

from moksha.api.hub import Consumer

class MokshaFeedConsumer(Consumer):

    topic = 'moksha.feeds'

    def __init__(self, hub):
        super(MokshaFeedConsumer, self).__init__(hub)
        self.feed_storage = Shove(config.get('feed_store', 'simple://'),
                                  config.get('feed_cache', 'simple://'),
                                  compress=True)

    def consume(self, message):
        self.log.debug('MokshaFeedConsumer.consume(%s)' % message)
        action = message['body'].get('action')
        if action and hasattr(self, action):
            getattr(self, action)(message['body'])
        else:
            # new feed found?
            pass

    def get_feed(self, message):
        url = message['key']
        feed = self.feed_storage[url][1]
        entries = []
        for entry in feed['entries']:
            entries.append({
                'title': entry.get('title'),
                'isLazy': False,
                'isFolder': False,
                'rootVisible': False,
                'key': "%s::%s" % (url, entry.get('title', '').replace(' ','')),
                })
        self.send_message(message['topic'], {
                'entries': entries,
                'key': url,
                'action': 'get_feed'
                })

    def get_entry(self, message, **kw):
        key = message['key']
        if '::' in key:
            url, title = key.split('::')
            for entry in self.feed_storage[url][1]['entries']:
                content = entry.get('content', entry.get('summary'))
                if isinstance(content, list):
                    if isinstance(content[0], dict) and \
                       content[0].has_key('value'):
                        content = content[0]['value']
                content = """
                    <blockquote><h3><a href="%s">%s</a></h3><br/>%s</blockquote>
                """ % (entry.get('link', url), entry['title'], content)
                if entry['title'].replace(' ', '') == to_unicode(title):
                    self.send_message(message['topic'], {
                            'content': content,
                            'action': 'get_entry'
                            })
                    return
            raise Exception("Cannot find entry by title: %s" % title)
        else:
            raise Exception("Invalid entry key: %r" % key)



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`moksha.widgets.feedtree` - A dynamic feed tree
====================================================

There are currently two implementations of this application, an `ajax` version
and a `live` version.  The ajax version makes a new request to our WSGI server
each time, where as the `live` implementation communicates over a persistent
Stomp-driven Orbited TCPSocket.   The live widget will automatically connect up
to a unique message topic that it uses to communicate with the Moksha Feed
Consumer in the Moksha Hub.  It also listens for changes in the feeds that
it is viewing.

.. moduleauthor:: Luke Macken <lmacken@redhat.com>
"""

from tg import config

import tw2.core as twc
from tw2.core import js_callback, JSLink, CSSLink
from tw2.jquery import jquery_js, jQuery
from tw2.jqplugins.dynatree import DynaTreeWidget
from tw2.jqplugins.cookies import jquery_cookies_js
from uuid import uuid4

from moksha.api.widgets.live import LiveWidget


class MokshaAjaxFeedTree(DynaTreeWidget):
    title = 'Moksha Ajax Feed Tree'
    options = {
        'rootVisible': True,
        'persist': True,
        'initAjax': {
            'url': '/apps/feeds/init_tree',
            'data': {'key': 'root'},
        },
        'onActivate': js_callback("""
            function(dtnode) {
              $('#TopPane').load('/apps/feeds/get_entries?key=' + dtnode.data.key.replace(/ /, ''));
            }
        """.replace('\n', '')),
    }


class MokshaAjaxFeedEntriesTree(DynaTreeWidget):
    options = {
        'rootVisible': False,
        'persist': True,
        'onActivate': js_callback("""function(dtnode) {
          $('#BottomPane').load('/apps/feeds/get_entry?key='+dtnode.data.key);
         }"""),
    }


class MokshaLiveFeedTree(DynaTreeWidget):
    title = 'Moksha Live Feed Tree'
    options = {
        'rootVisible': True,
        'persist': True,
        'fx': {'height': 'toggle', 'duration': 200},
        'initAjax': {
            'url': '/apps/feeds/init_tree',
            'data': {'key': 'root'}
        },
        'onActivate': js_callback("function(dtnode) { moksha.send_message('moksha.feeds', {action: 'get_feed', 'key': dtnode.data.key, topic: moksha_feed_topic}); }")
    }

    def prepare(self):
        self.topic = str(uuid4())
        super(MokshaLiveFeedTree, self).prepare()


class MokshaLiveFeedEntriesTree(DynaTreeWidget):
    options = {
        'rootVisible': False,
        'persist': True,
        'onActivate': js_callback("""
            function(dtnode) {
                moksha.send_message('moksha.feeds', {
                    'action': 'get_entry',
                    'key': dtnode.data.key,
                     topic: moksha_feed_topic
                 });

                /* Unsubscribe from current feed, subscribe to new one */
            }
        """.replace('\n', '')),
    }


## Load our feed tree widgets.
feedtree_engine = config.get('mdemos.feeds.feedtree.engine', 'live')
if feedtree_engine == 'live':   # Live widgets
    feed_tree = MokshaLiveFeedTree(id='feed_tree')
    feed_entries_tree = MokshaLiveFeedEntriesTree(id='feed_entries_tree')
elif feedtree_engine == 'ajax': # Ajax widgets
    feed_tree = MokshaAjaxFeedTree(id='feed_tree')
    feed_entries_tree = MokshaAjaxFeedEntriesTree(id='feed_entries_tree')

splitter_js = JSLink(filename='static/splitter.js',
                     resources=[jquery_js],
                     modname=__name__)

splitter_css = CSSLink(filename='static/main.css',
                       media='all',
                       modname=__name__)


class MokshaFeedReaderWidget(LiveWidget):
    name = 'Moksha Feed Reader'
    topic = 'moksha.feeds' # will get replaced by a unique uuid at render-time
    template = 'mako:mdemos.feeds.widgets.templates.feedreader'

    # Children
    feed_tree = twc.Variable(default=feed_tree)
    feed_entries_tree = twc.Variable(default=feed_entries_tree)

    resources = [
        jquery_cookies_js,
        splitter_js,
        splitter_css,
    ]

    container_options = {
        'top': 50,
        'left': 50,
        'height': 600,
        'width': 890,
        'icon': 'browser.png',
    }

    onmessage = """
      if (json.action == 'get_feed') {
          var tree = $("#moksha_feedreader_feed_entries_tree").dynatree("getRoot");
          tree.removeChildren();
          tree.append(json.entries);
      } else if (json.action == 'get_entry') {
          $('#BottomPane').html(json.content);
      } else if (json.action == 'new_entry') {
          /* TODO */
          moksha.debug('new_entry!');
      }
    """

    def prepare(self):
        self.topic = str(uuid4())
        super(MokshaFeedReaderWidget, self).prepare()
        self.add_call(jQuery('#' + self.id).splitter({
            'splitVertical': True,
            'outline': True,
            'sizeLeft': True,
            'anchorToWindow': True,
            'accessKey': "I",
            }))
        self.add_call(jQuery('#RightPane').splitter({
            'splitHorizontal': True,
            'sizeTop': True,
            'accessKey': "H",
            }))

moksha_feedreader = MokshaFeedReaderWidget(id='moksha_feedreader')



# This file is part of Moksha.
# Copyright (C) 2008-2010  Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Moksha Feed Tree Controller

.. moduleauthor:: Luke Macken <lmacken@redhat.com>
"""

import moksha
import moksha.utils
import logging

from tg import expose, tmpl_context
from pylons import cache
from orbited import json
from kitchen.text.converters import to_unicode

from moksha.lib.base import Controller
from mdemos.feeds.widgets import feed_entries_tree, moksha_feedreader

from tg import expose, validate
from formencode import validators

log = logging.getLogger('moksha.hub')

class FeedController(Controller):

    @expose('mako:mdemos.feeds.templates.index')
    @validate({'name': validators.UnicodeString()})
    def index(self, name='world', *args, **kw):
        return dict(name=name)

    @expose()
    def init_tree(self, key, fresh=False, **kw):
        c = cache.get_cache('feeds')
        if fresh:
            return self._get_feed_titles(fresh=fresh)
        else:
            return c.get_value(key='feed_titles',
                               createfunc=self._get_feed_titles,
                               expiretime=3600)

    def _get_feed_titles(self, fresh=False):
        results = []
        for feed in moksha.utils.feed_storage.keys():
            if not feed:
                raise Exception('None feed?!')
            if fresh:
                print "Getting fresh data"
                feed_data = moksha.utils.feed_cache.fetch(feed)
            else:
                print "Getting cached data"
                timestamp, feed_data = moksha.utils.feed_storage[feed]
            if not feed_data:
                log.debug("no feed_data, refetching")
                #feed_data = moksha.utils.feed_cache.fetch(feed)
                #if not feed_data:
                #    log.debug("NO FEED DATA AFTER FRESH FETCH!!!!")
                continue
            channel = feed_data.get('channel')
            if not channel:
                continue
            title = channel.get('title')
            results.append({
                'title': title,
                'key': feed,
                'isLazy': False,
                'isFolder': True,
                })
        return json.encode(results)

    @expose()
    def get_feed(self, key, *args, **kw):
        if '::' in key:
            url, title = key.split('::')
            for entry in moksha.utils.feed_storage[url][1]['entries']:
                content = entry.get('content', entry.get('summary'))
                content = '<span style="line-height:100%%;">%s</span>' % content
                if entry['title'] == title:
                    return json.encode([{'title': content, 'isLazy': False}])
            raise Exception("Cannot find entry by title: %s" % title)

        feed = moksha.utils.feed_storage[key][1]
        entries = []
        for entry in feed['entries']:
            entries.append({
                'title': entry.get('title'),
                'isLazy': True,
                'isFolder': True,
                'key': "%s::%s" % (key, entry.get('title')),
                })
        return json.encode(entries)

    @expose('mako:moksha.templates.widget')
    def get_entries(self, key, **kw):
        tmpl_context.widget = feed_entries_tree
        feed = moksha.utils.feed_storage[key][1]
        entries = []
        for entry in feed['entries']:
            entries.append({
                'title': entry.get('title'),
                'isLazy': False,
                'isFolder': False,
                'rootVisible': False,
                'key': "%s::%s" % (key, entry.get('title','').replace(' ','')),
                })
        return dict(options={'tree_children': entries})

    @expose()
    def get_entry(self, key, **kw):
        if '::' in key:
            url, title = key.split('::')
            for entry in moksha.utils.feed_storage[url][1]['entries']:
                content = entry.get('content', entry.get('summary'))
                if isinstance(content, list):
                    if isinstance(content[0], dict) and \
                       content[0].has_key('value'):
                        content = content[0]['value']
                content = """
                    <blockquote><h3><a href="%s">%s</a></h3><br/>%s</blockquote>
                """ % (entry.get('link', url), entry['title'], content)
                if entry['title'].replace(' ', '') == to_unicode(title):
                    return content
            raise Exception("Cannot find entry by title: %s" % title)
        else:
            raise Exception("Invalid entry key: %r" % key)



# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup


setup(
    name="mdemos.feeds",
    version="0.1.1a",
    url="http://moksha.fedorahosted.org",
    description="Moksha Feeds App",
    license="ASL 2.0",
    long_description="",
    author="Luke Macken",
    author_email="lmacken@redhat.com",
    packages=['mdemos', 'mdemos.feeds'],
    namespace_packages=[
        'mdemos',
    ],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "moksha>=0.7.0a",
        "tw2.jqplugins.dynatree",
        "tw2.jqplugins.cookies",
    ],
    entry_points={
        'moksha.producer': (
            'feeds = mdemos.feeds.streams:MokshaFeedStream',
        ),
        'moksha.consumer': (
            'feeds = mdemos.feeds.consumer:MokshaFeedConsumer'
        ),
        'moksha.application': (
            'feeds = mdemos.feeds.controllers:FeedController'
        ),
        'moksha.widget': (
            'feeds = mdemos.feeds.widgets:moksha_feedreader'
        ),
    }
)
